// Visualization interface to Graphviz dotty(1) -*- c++ -*-

#ifndef DOTTY_H_
# define DOTTY_H_
# ifdef __GNUC__
#  pragma interface
# endif // __GNUC__

# include "typedefs.h"

/** @file Dotty.h
 * Visualization interface to Graphviz dotty(1)
 */

/** Graphviz dotty(1) interface */
class Dotty
{
protected:
  /** Constructor */
  Dotty ();
private:
  /** Copy constructor */
  Dotty (const class Dotty& old);
  /** Assignment operator */
  class Dotty& operator= (const class Dotty& old);
public:
  /** Destructor */
  ~Dotty ();

  /** Set the name of the visualization program */
  static void useVisual (const char* name);
  /** Start the visualization process
   * @return		pointer to the visualization object; NULL on failure
   */
  static class Dotty* startVisual ();
  /** Stop the visualization process */
  static void stopVisual ();
private:
  /** Clear the visualization variables */
  void clearVisual ();
public:

  /** Dump the syntax tree of the model
   * @param net		the model being dumped
   * @param adding	flag: add to existing graph
   */
  void dump (const class Net& net,
	     bool adding) const;

  /** Display a state
   * @param graph	the reachability graph
   * @param cgraph	the graph of strongly connected components (optional)
   * @param state	number of the state to be displayed
   * @param adding	flag: add to existing graph
   */
  void show (const class Graph& graph,
	     const class ComponentGraph* cgraph,
	     card_t state, bool adding) const;

  /** Display successor arcs
   * @param reporter	the reachability graph interface
   * @param cgraph	the graph of strongly connected components (optional)
   * @param state	number of the originating state
   * @param seq		flag: fully expand deterministic sequences
   * @param adding	flag: add to existing graph
   */
  void succ (class GraphReporter& reporter,
	     const class ComponentGraph* cgraph,
	     card_t state, bool seq, bool adding) const;

  /** Display predecessor arcs
   * @param graph	the reachability graph
   * @param cgraph	the graph of strongly connected components (optional)
   * @param state	number of the target state
   * @param seq		flag: fully expand deterministic sequences
   * @param adding	flag: add to existing graph
   */
  void pred (const class Graph& graph,
	     const class ComponentGraph* cgraph,
	     card_t state, bool seq, bool adding) const;

  /** Display a graph of strongly connected components
   * @param cgraph	the graph of strongly connected components
   * @param adding	flag: add to existing graph
   */
  void components (const class ComponentGraph& cgraph,
		   bool adding) const;

  /** Display a strongly connected component
   * @param graph	the reachability graph
   * @param cgraph	the graph of strongly connected components
   * @param comp	number of the component
   * @param cond	condition for picking the states in the component
   * @param adding	flag: add to existing graph
   */
  void component (const class Graph& graph,
		  const class ComponentGraph& cgraph,
		  card_t comp,
		  const class Expression* cond,
		  bool adding) const;

  /** Display the succeeding components of a strongly connected component
   * @param cgraph	the graph of strongly connected components
   * @param comp	number of the component
   * @param adding	flag: add to existing graph
   */
  void comp_succ (const class ComponentGraph& cgraph,
		  card_t comp, bool adding) const;

  /** Display the preceding components of a strongly connected component
   * @param cgraph	the graph of strongly connected components
   * @param comp	number of the component
   * @param adding	flag: add to existing graph
   */
  void comp_pred (const class ComponentGraph& cgraph,
		  card_t comp, bool adding) const;

  /** Display a path in a reachability graph
   * @param graph	the reachability graph
   * @param cgraph	the graph of strongly connected components (optional)
   * @param path	state numbers on the path
   * @param adding	flag: add to existing graph
   * @param reverse	flag: reverse the arcs
   */
  void show (const class Graph& graph,
	     const class ComponentGraph* cgraph,
	     const card_t* path,
	     bool adding,
	     bool reverse) const;

  /** @name The lower-level interface for displaying counterexample paths */
  /*@{*/
  /** Display the graph prologue
   * @param adding	flag: add to existing graph
   */
  void displayPrologue (bool adding) const;
  /** Display the graph epilogue */
  void displayEpilogue () const;
  /** Display a node in a counterexample path
   * @param m		the marking to be displayed
   */
  void displayMarking (const class GlobalMarking& m) const;
  /** Display an edge in a counterexample path
   * @param transition	the transition
   * @param valuation	the valuation
   * @param m		the destination node
   */
  void displayEdge (const class Transition& transition,
		    const class Valuation& valuation,
		    const class GlobalMarking& m) const;
  /*@}*/

private:
  /** The singleton object */
  static class Dotty theDotty;
  /** The output stream */
  class Printer* myOut;
};

#endif // DOTTY_H_
